'use strict';

var fs = require('fs');
var htmlMinifier = require('html-minifier');
var nodeHtmlParser = require('node-html-parser');
var path = require('path');

const getChildElement = (node, tag, append = true) => {
    let child = node.querySelector(tag);
    if (!child) {
        child = new nodeHtmlParser.HTMLElement(tag, {});
        if (append) {
            node.appendChild(child);
        }
        else {
            node.childNodes.unshift(child);
        }
    }
    return child;
};
const addNewLine = (node) => node.appendChild(new nodeHtmlParser.TextNode('\n  '));
var ExternalPosition;
(function (ExternalPosition) {
    ExternalPosition["before"] = "before";
    ExternalPosition["after"] = "after";
})(ExternalPosition || (ExternalPosition = {}));
var Crossorigin;
(function (Crossorigin) {
    Crossorigin["anonymous"] = "anonymous";
    Crossorigin["usecredentials"] = "use-credentials";
})(Crossorigin || (Crossorigin = {}));
const normalizePreload = (preload) => {
    if (!preload) {
        preload = [];
    }
    if (preload instanceof Array) {
        preload = new Set(preload);
    }
    return preload;
};
const extensionToType = (ext) => {
    switch (ext) {
        case 'css': return 'style';
        case 'js': return 'script';
        default: return null;
    }
};
const isChunk = (item) => {
    return !item.isAsset;
};
const bundleReducer = (prev, cur) => {
    if (isChunk(cur)) {
        // Use full name with possible hash and without extension to process
        // possible CSS files and other assets with same name of entry
        const { name } = path.parse(cur.fileName);
        if (cur.isEntry) {
            prev.entries[name] = cur.name;
        }
        else if (cur.isDynamicEntry) {
            prev.dynamicEntries[name] = cur.name;
        }
    }
    return prev;
};
const formatSupportsModules = (f) => (f === 'es'
    || f === 'esm'
    || f === 'module');
const checkEnum = (enumobj, val) => (!val || Object.values(enumobj).includes(val));
const injectCSSandJSFactory = (head, body, modules) => {
    const typeModule = modules ? 'type="module" ' : '';
    return (fileName, type, pos, crossorigin) => {
        const cors = crossorigin ? `crossorigin="${crossorigin}" ` : '';
        if (type === "css" /* css */) {
            const parent = pos === "body" /* body */ ? body : head;
            addNewLine(parent);
            parent.appendChild(new nodeHtmlParser.HTMLElement('link', {}, `rel="stylesheet" ${cors}href="${fileName}"`));
        }
        else {
            const parent = pos === "head" /* head */ ? head : body;
            addNewLine(parent);
            parent.appendChild(new nodeHtmlParser.HTMLElement('script', {}, `${typeModule}${cors}src="${fileName}"`));
        }
    };
};
const extrenalsProcessorFactory = (injectCSSandJS, externals) => {
    if (!externals) {
        // tslint:disable-next-line: no-empty
        return (_pos) => { };
    }
    return (processPos) => {
        for (const { pos, file, type, crossorigin } of externals) {
            if (pos === processPos) {
                injectCSSandJS(file, type || path.extname(file).slice(1), undefined, crossorigin);
            }
        }
    };
};
var index = ({ template, file, inject, title, favicon, meta, externals, preload, modules, minify: minifyOptions, onlinePath = '', ...options }) => ({
    name: 'html2',
    buildStart() {
        const isHTML = /^.*<html>[\s\S]*<\/html>\s*$/i.test(template);
        if (isHTML) {
            if (!file) {
                this.error('When `template` is an HTML string the `file` option must be defined');
            }
        }
        else {
            if (fs.existsSync(template) && fs.lstatSync(template).isFile()) {
                this.addWatchFile(template);
            }
            else {
                this.error('`template` must be an HTML or a file path');
            }
        }
        this.cache.set("isHTML" /* isHTML */, isHTML);
        if (favicon && !(fs.existsSync(favicon) && fs.lstatSync(favicon).isFile())) {
            this.error('The provided favicon file does\'t exist');
        }
        if (typeof inject === 'string' && !(inject === "head" /* head */ || inject === "body" /* body */)) {
            this.error('Invalid inject argument: ' + inject);
        }
        if (externals) {
            for (const { pos, crossorigin } of externals) {
                if (!checkEnum(ExternalPosition, pos)) {
                    this.error('Invalid position for the extrenal: ' + pos);
                }
                if (!checkEnum(Crossorigin, crossorigin)) {
                    this.error('Invalid crossorigin argument for the extrenal: ' + crossorigin);
                }
            }
        }
        const typeofmodules = typeof modules;
        if (typeofmodules !== 'boolean' && typeofmodules !== 'undefined') {
            this.error('Invalid `modules` argument: ' + JSON.stringify(modules));
        }
        Object.keys(options).forEach(o => this.warn(`Ignoring unknown option "${o}"`));
    },
    outputOptions({ dir, file: bundleFile, format }) {
        if (!file) {
            let distDir = process.cwd();
            if (dir) {
                distDir = path.resolve(distDir, dir);
            }
            else if (bundleFile) {
                const bundleDir = path.dirname(bundleFile);
                distDir = path.isAbsolute(bundleDir) ? bundleDir : path.resolve(distDir, bundleDir);
            }
            // Template is always a file path
            file = path.resolve(distDir, path.basename(template));
            if (file === path.resolve(template)) {
                this.error('Could\'t write the generated HTML to the source template, define one of the options: `file`, `output.file` or `output.dir`');
            }
        }
        if (modules && !formatSupportsModules(format)) {
            this.error(`The modules option is set to true but the output.format is ${format}, \
consider to use the esm format or switch off the option`);
        }
        return null;
    },
    generateBundle(_output, bundle) {
        const data = this.cache.get("isHTML" /* isHTML */)
            ? template
            : fs.readFileSync(template).toString();
        const doc = nodeHtmlParser.parse(data, {
            pre: true,
            script: true,
            style: true,
        });
        if (!doc.valid) {
            this.error('Error parsing template');
        }
        const html = doc.querySelector('html');
        if (!html) {
            this.error('The input template doesn\'t contain the `html` tag');
        }
        const head = getChildElement(html, 'head', false);
        const body = getChildElement(html, 'body');
        if (meta) {
            const nodes = head.querySelectorAll('meta');
            Object.entries(meta).forEach(([name, content]) => {
                const oldMeta = nodes.find(n => n.attributes.name === name);
                const newMeta = new nodeHtmlParser.HTMLElement('meta', {}, `name="${name}" content="${content}"`);
                if (oldMeta) {
                    head.exchangeChild(oldMeta, newMeta);
                }
                else {
                    addNewLine(head);
                    head.appendChild(newMeta);
                }
            });
        }
        if (title) {
            let node = head.querySelector('title');
            if (!node) {
                addNewLine(head);
                node = new nodeHtmlParser.HTMLElement('title', {});
                head.appendChild(node);
            }
            node.set_content(title);
        }
        if (favicon) {
            const nodes = head.querySelectorAll('link');
            const rel = 'shortcut icon';
            const oldLink = nodes.find(n => n.attributes.rel === rel);
            const fileName = path.basename(favicon);
            const newLink = new nodeHtmlParser.HTMLElement('link', {}, `rel="${rel}" href="${fileName}"`);
            if (oldLink) {
                head.exchangeChild(oldLink, newLink);
            }
            else {
                addNewLine(head);
                head.appendChild(newLink);
            }
            bundle[fileName] = {
                fileName,
                isAsset: true,
                source: fs.readFileSync(favicon),
                type: 'asset',
            };
        }
        const injectCSSandJS = injectCSSandJSFactory(head, body, modules);
        const processExternals = extrenalsProcessorFactory(injectCSSandJS, externals);
        // Inject externals before
        processExternals(ExternalPosition.before);
        // Inject generated files
        if (inject !== false) {
            const files = Object.values(bundle);
            // First of all get entries
            const { entries, dynamicEntries } = files.reduce(bundleReducer, {
                dynamicEntries: {},
                entries: {},
            });
            // Now process all files and inject only entries and preload files
            preload = normalizePreload(preload);
            files.forEach(({ fileName }) => {
                const { name, ext } = path.parse(fileName);
                const injectType = ext.slice(1);
                if (name in entries) {
                    injectCSSandJS(`${onlinePath}/${fileName}`, injectType, inject);
                }
                else if (name in dynamicEntries && preload.has(dynamicEntries[name])) {
                    const linkType = extensionToType(injectType);
                    if (linkType) {
                        addNewLine(head);
                        head.appendChild(new nodeHtmlParser.HTMLElement('link', {}, `rel="preload" href="${onlinePath}/${fileName}" as="${linkType}"`));
                    }
                }
            });
        }
        // Inject externals after
        processExternals(ExternalPosition.after);
        let source = '<!doctype html>\n' + doc.toString();
        if (minifyOptions) {
            source = htmlMinifier.minify(source, minifyOptions);
        }
        bundle[file] = {
            fileName: file,
            isAsset: true,
            source,
            type: 'asset',
        };
    },
});

module.exports = index;
